#  Copyright © 2025 Bentley Systems, Incorporated
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from __future__ import annotations

import json
from collections.abc import Callable, Iterator, Mapping, Sequence
from types import MappingProxyType
from typing import Any, Generic, TypeVar, overload
from uuid import UUID

try:
    import jmespath
    from jmespath.exceptions import JMESPathError
    from jmespath.parser import ParsedResult as UpstreamParsedResult
    from jmespath.visitor import Options
except ImportError:
    raise ImportError(
        "The 'jmespath' package is required for evo.json.jmespath. "
        "Please install 'evo-sdk-common' with the 'jmespath' extra."
    ) from None

__all__ = [
    "JMESPathArrayProxy",
    "JMESPathError",
    "JMESPathObjectProxy",
    "Options",
    "ParsedResult",
    "compile",
    "proxy",
    "search",
]

T = TypeVar("T")


class _JMESPathViewEncoder(json.JSONEncoder):
    def default(self, obj: Any) -> Any:
        if isinstance(obj, MappingProxyType):
            return dict(obj)

        if isinstance(obj, UUID):
            return str(obj)

        return super().default(obj)


class _JMESPathProxyMixin(Generic[T]):
    def __init__(self, data: T) -> None:
        self._data = data

    @property
    def raw(self) -> T:
        return self._data

    def search(self, expression: str) -> Any:
        """Search the proxied data with a JMESPath expression.

        :param expression: The JMESPath expression to compile and search.

        :return: The result of the search, as JMESArrayProxy, JMESObjectProxy, or a primitive type.

        :raises JMESPathError: If the expression is invalid.
        """
        return search(expression, self._data)

    def json_dumps(
        self,
        /,
        skipkeys: bool = False,
        ensure_ascii: bool = False,
        check_circular: bool = True,
        allow_nan: bool = True,
        cls: type[json.JSONEncoder] | None = None,
        indent: int | str | None = None,
        separators: tuple[str, str] | None = None,
        default: Callable[[Any], Any] | None = None,
        sort_keys: bool = False,
        **kwargs: Any,
    ) -> str:
        """Serialize the contained data to a JSON string.

        This method is the same as json.dumps, but uses a custom encoder by default that
        handles MappingProxyType and UUID.

        All parameters are the same as for json.dumps, except that cls defaults to
        our custom encoder.
        """
        if cls is None:
            # Use our own encoder that handles MappingProxyType and UUID.
            cls = _JMESPathViewEncoder

        return json.dumps(
            self._data,
            skipkeys=skipkeys,
            ensure_ascii=ensure_ascii,
            check_circular=check_circular,
            allow_nan=allow_nan,
            cls=cls,
            indent=indent,
            separators=separators,
            default=default,
            sort_keys=sort_keys,
            **kwargs,
        )

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self.json_dumps(indent=2)})"

    def __eq__(self, other: Any) -> bool:
        if isinstance(other, _JMESPathProxyMixin):
            return self.raw == other.raw
        else:
            return self.raw == other


class JMESPathArrayProxy(Generic[T], _JMESPathProxyMixin[Sequence[T]], Sequence[T]):
    def __getitem__(self, index: int | str) -> Any:
        if isinstance(index, int):
            return proxy(self.raw[index])
        else:
            return self.search(index)

    def __len__(self) -> int:
        return len(self._data)

    def __iter__(self) -> Iterator[Any]:
        for item in self.raw:
            yield proxy(item)


class JMESPathObjectProxy(Generic[T], _JMESPathProxyMixin[Mapping[str, T]], Mapping[str, T]):
    def __getitem__(self, key: str) -> Any:
        return self.search(key)

    def __iter__(self) -> Iterator[str]:
        return iter(self.raw)

    def __len__(self) -> int:
        return len(self.raw)


@overload
def proxy(value: Mapping[str, T]) -> JMESPathObjectProxy[T]: ...


@overload
def proxy(value: Sequence[T]) -> JMESPathArrayProxy[T]: ...


@overload
def proxy(value: T) -> T: ...


def proxy(value: Any) -> Any:
    """Convert a JSON-like value into a JMESPath proxy type if applicable."""
    if isinstance(value, Mapping):
        return JMESPathObjectProxy(value)
    elif isinstance(value, Sequence) and not isinstance(value, str):
        return JMESPathArrayProxy(value)
    else:
        return value


class ParsedResult(UpstreamParsedResult):
    def search(self, value: Any, options: Options | None = None) -> Any:
        return proxy(super().search(value, options))


def compile(expression: str) -> ParsedResult:
    """Thin wrapper around jmespath.compile to return our own version of ParsedResult.

    :param expression: The JMESPath expression to compile.

    :return: A ParsedResult instance.

    :raises JMESPathError: If the expression is invalid.
    """
    result = jmespath.compile(expression)
    return ParsedResult(result.expression, result.parsed)


def search(expression: str, data: Any, options: Options | None = None) -> Any:
    """Reimplementation of jmespath.search that returns our proxy types for JSON results.

    :param expression: The JMESPath expression to compile.
    :param data: The data to search.
    :param options: Optional jmespath Options.

    :return: The result of the search, as JMESArrayProxy, JMESObjectProxy, or a primitive type.

    :raises JMESPathError: If the expression is invalid.
    """
    return compile(expression).search(data, options=options)



#  Copyright © 2025 Bentley Systems, Incorporated
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

"""Evo Client Common Test Tools

The contents of this module are utilities that make it easy to test service client libraries. These utilities should
only be used for unit tests.
"""

import sys
import warnings

from .consts import ACCESS_TOKEN, BASE_URL, HUB, ORG, WORKSPACE_ID
from .http import (
    AbstractTestRequestHandler,
    MockResponse,
    TestAuthorizer,
    TestHTTPHeaderDict,
    TestTransport,
    TestWithConnector,
)
from .io import (
    DownloadRequestHandler,
    MultiDownloadRequestHandler,
    StorageDestinationRequestHandler,
    TestWithDownloadHandler,
    TestWithUploadHandler,
    UrlGenerator,
)
from .storage import TestWithStorage
from .utils import long_test, utc_datetime, utc_time

if "pytest" not in sys.modules:
    # Issue a warning whenever this module is imported.
    warnings.warn(__doc__)

__all__ = [
    "ACCESS_TOKEN",
    "BASE_URL",
    "HUB",
    "ORG",
    "WORKSPACE_ID",
    "AbstractTestRequestHandler",
    "DownloadRequestHandler",
    "MockResponse",
    "MultiDownloadRequestHandler",
    "StorageDestinationRequestHandler",
    "TestAuthorizer",
    "TestHTTPHeaderDict",
    "TestTransport",
    "TestWithConnector",
    "TestWithDownloadHandler",
    "TestWithStorage",
    "TestWithUploadHandler",
    "UrlGenerator",
    "long_test",
    "utc_datetime",
    "utc_time",
]



#  Copyright © 2025 Bentley Systems, Incorporated
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from abc import ABC, abstractmethod
from collections.abc import Iterator
from contextlib import contextmanager

from evo import logging

from .._types import PathLike
from ..interfaces import IFeedback, ITransport
from ..utils import NoFeedback, Retry
from .storage import StorageDestination

logger = logging.getLogger("io.upload")


class Upload(ABC):
    """A base class for referencing binary data that needs to be uploaded.

    Each SDK that accepts binary data from an API is expected to extend this class, providing the specific implementation
    for generating upload URLs.
    """

    @property
    @abstractmethod
    def label(self) -> str:
        """The label of the resource to be uploaded."""
        ...

    @abstractmethod
    async def get_upload_url(self) -> str:
        """Generate a URL that will be used to upload the resource.

        This method may be called multiple times to generate a new URL if the last URL expires.

        :returns: The upload URL.

        :raises DataExistsError: if the resource already exists.
        """
        ...

    @contextmanager
    def _uploading(self) -> Iterator[None]:
        """A context manager that logs the start and end of the upload process."""
        logger.debug(f"Attempting to upload data (label: {self.label})")
        yield
        logger.debug(f"Data uploaded successfully (label: {self.label})")

    async def upload_from_path(
        self,
        filename: PathLike,
        transport: ITransport,
        max_workers: int | None = None,
        retry: Retry | None = None,
        fb: IFeedback = NoFeedback,
    ) -> None:
        """Upload a file with the given filename.

        :param filename: file to upload
        :param transport: The transport to use for the API calls.
        :param max_workers: The maximum number of concurrent connections to use for data transfer. If None, the default
            is `min(32, os.cpu_count() + 4)`
        :param retry: A Retry object with a wait strategy. If None, a default Retry is created. If a chunk is successfully
            transferred the attempt counter will be reset.
        :param fb: feedback to track the upload, by tracking reads from the file only

        :raises ValueError: if the file to upload does not exist
        :raises ChunkedIOException: if a non-recoverable exception occurred.
        :raises RetryError: if the maximum number of consecutive attempts have failed. RetryError is a subclass
            of ChunkedIOException.
        """
        with self._uploading():
            await StorageDestination.upload_file(
                filename=filename,
                url_generator=self.get_upload_url,
                transport=transport,
                max_workers=max_workers,
                retry=retry,
                fb=fb,
            )



#  Copyright © 2025 Bentley Systems, Incorporated
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from pathlib import Path
from typing import TypeAlias

import ipywidgets as widgets

from evo.common.utils import Cache

from . import assets
from ._consts import DEFAULT_CACHE_LOCATION

FileName: TypeAlias = str | Path


def init_cache(cache_location: FileName = DEFAULT_CACHE_LOCATION) -> Cache:
    """Initialise the storage location for the notebook environment.

    Configures the cache location and creates a `.gitignore` file in the root of the cache directory.

    :param cache_location: The location for the cache directory.

    :returns: A Cache instance.
    """
    cache = Cache(cache_location, mkdir=True)
    ignorefile = cache.root / ".gitignore"
    ignorefile.write_text("*\n")
    return cache


def build_img_widget(filename: str) -> widgets.Image:
    image = assets.get(filename).read_bytes()
    return widgets.Image(
        value=image,
        format="png",
        layout=widgets.Layout(max_height="26px", margin="3px", align_self="center"),
    )


def build_button_widget(text: str) -> widgets.Button:
    widget = widgets.Button(
        description=text,
        button_style="info",
        layout=widgets.Layout(margin="5px 5px 5px 5px", align_self="center"),
    )
    widget.style.button_color = "#265C7F"
    return widget



#  Copyright © 2025 Bentley Systems, Incorporated
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
"""
Workspaces API
=============

The Workspaces API enables users to organize, maintain, and store project data, but does not use or process the data. The workspace APIs allow you to manage:
- Workspaces
- User roles within workspaces
- Workspace thumbnails

There are three pre-defined roles within workspaces:

- Owner: can perform all actions in the workspace
- Editor: can perform all actions excluding deleting of a workspace
- Viewer: can view the workspace

These user roles can be assigned to users in a workspace. Once a role has been assigned it can be replaced or removed.
Users can also retrieve user roles, the role of a particular user, and their own role if applicable.
For more information on using the Workspaces API, see the [Workspaces API overview](https://developer.seequent.com/docs/guides/workspaces/), or the API references here.


This code is generated from the OpenAPI specification for Workspaces API.
API version: 1.0
"""

# Import endpoint apis.
from .api import (
    AdminApi,
    DiscoveryApi,
    FoldersApi,
    GeneralApi,
    HubsApi,
    InstancesApi,
    InstanceUsersApi,
    LicenseAccessApi,
    ThumbnailsApi,
    TokenApi,
    WorkspacesApi,
)

__all__ = [
    "AdminApi",
    "DiscoveryApi",
    "FoldersApi",
    "GeneralApi",
    "HubsApi",
    "InstanceUsersApi",
    "InstancesApi",
    "LicenseAccessApi",
    "ThumbnailsApi",
    "TokenApi",
    "WorkspacesApi",
]



#  Copyright © 2025 Bentley Systems, Incorporated
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from collections.abc import Iterator, Sequence
from threading import Lock
from typing import TypeVar

from ..interfaces import IFeedback

__all__ = [
    "NoFeedback",
    "PartialFeedback",
    "iter_with_fb",
]

_N_DIGITS = 4  # Let's maintain a user-friendly number of dp.


class _NoFeedback(IFeedback):
    def progress(self, progress: float, message: str | None = None) -> None:
        pass


NoFeedback = _NoFeedback()
"""A default feedback object that does nothing. Use this when no feedback is needed."""


class PartialFeedback(IFeedback):
    """A wrapper for IFeedback objects that subdivides the feedback range."""

    def __init__(self, parent: IFeedback, start: float, end: float) -> None:
        """
        :param parent: The parent feedback object to subdivide.
        :param start: The start of the partial feedback range.
        :param end: The end of the partial feedback range.
        """
        self.__offset = start
        self.__factor = end - start

        self.__lock = Lock()
        self.__parent = parent

    def progress(self, progress: float, message: str | None = None) -> None:
        """Update the progress of the feedback object.

        :param progress: The progress value, between 0 and 1.
        :param message: An optional message to display.
        """
        partial_progress = round(self.__offset + (progress * self.__factor), ndigits=_N_DIGITS)
        with self.__lock:
            self.__parent.progress(partial_progress, message)


T = TypeVar("T")


def iter_with_fb(elements: Sequence[T], feedback: IFeedback | None = None) -> Iterator[tuple[T, IFeedback]]:
    """Iterate over a sequence of elements, dividing feedback uniformly throughout the range.

    :param elements: The sequence of elements to iterate over.
    :param feedback: A feedback object to subdivide.

    :yields: A tuple containing a sequence element, and a new feedback object for that element.
    """
    if len(elements) == 0:
        return

    fb_part_size = 1 / len(elements)
    for i, element in enumerate(elements):
        if feedback is None:
            yield element, NoFeedback
        else:
            start = round(i * fb_part_size, ndigits=_N_DIGITS)
            end = round((i + 1) * fb_part_size, ndigits=_N_DIGITS)
            yield element, PartialFeedback(feedback, start, end)
            feedback.progress(end)



#  Copyright © 2025 Bentley Systems, Incorporated
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from abc import ABC, abstractmethod
from collections.abc import Iterator
from contextlib import contextmanager
from pathlib import Path
from typing import Generic, TypeVar

from evo import logging

from .._types import PathLike
from ..data import ResourceMetadata
from ..interfaces import ICache, IFeedback, ITransport
from ..utils import NoFeedback, Retry
from .http import HTTPSource

logger = logging.getLogger("io.download")

T = TypeVar("T", bound=ResourceMetadata)


class Download(ABC, Generic[T]):
    """A base class for referencing binary data that needs to be downloaded.

    Each SDK that provides binary data from an API is expected to extend this class, providing the specific implementation
    for generating download URLs and caching the downloaded data.
    """

    @property
    @abstractmethod
    def label(self) -> str:
        """The label of the resource to be downloaded."""
        ...

    @property
    @abstractmethod
    def metadata(self) -> T:
        """The metadata of the resource to be downloaded."""
        ...

    @abstractmethod
    def _get_cache_location(self, cache: ICache) -> Path:
        """Generate the cache location for the resource to be downloaded.

        :param cache: The cache to resolve the cache location.

        :returns: The cache location.
        """
        ...

    @abstractmethod
    async def get_download_url(self) -> str:
        """Generate a URL that will be used to download the resource.

        This method may be called multiple times to generate a new URL if the last URL expires.

        :returns: The download URL.
        """
        ...

    @contextmanager
    def _downloading(self) -> Iterator[None]:
        """A context manager that logs the start and end of the download process."""
        logger.debug(f"Attempting to download data (label: {self.label})")
        yield
        logger.debug(f"Data downloaded successfully (label: {self.label})")

    async def download_to_path(
        self,
        filename: PathLike,
        transport: ITransport,
        max_workers: int | None = None,
        retry: Retry | None = None,
        overwrite: bool = False,
        fb: IFeedback = NoFeedback,
    ) -> None:
        """Download an HTTP resource to a file with the given filename.

        The url generator MUST generate a resource URL that can be used to access the required resource. The URL
        generator may be called again if the last URL expires (unless Retry is initialised with max_attempts == 0).

        :param filename: file to download to
        :param transport: The transport to use for the API calls.
        :param max_workers: The maximum number of concurrent connections to use for data transfer. If None, the default
            is `min(32, os.cpu_count() + 4)`
        :param retry: A Retry object with a wait strategy. If None, a default Retry is created. If a chunk is successfully
            transferred the attempt counter will be reset.
        :param overwrite: whether to overwrite an existing local file
        :param fb: feedback to track the download, by tracking writes to the file

        :raises FileNameTooLongError: If the filename is too long.
        :raises ValueError: if the file already exists and overwrite is False
        :raises ChunkedIOException: if a non-recoverable exception occurred.
        :raises RetryError: if the maximum number of consecutive attempts have failed.
        """
        with self._downloading():
            await HTTPSource.download_file(
                filename=filename,
                url_generator=self.get_download_url,
                transport=transport,
                max_workers=max_workers,
                retry=retry,
                overwrite=overwrite,
                fb=fb,
            )

    async def download_to_cache(
        self,
        cache: ICache,
        transport: ITransport,
        max_workers: int | None = None,
        retry: Retry | None = None,
        overwrite: bool = False,
        fb: IFeedback = NoFeedback,
    ) -> Path:
        """Download an HTTP resource to a file with the given filename.

        The url generator MUST generate a resource URL that can be used to access the required resource. The URL
        generator may be called again if the last URL expires (unless Retry is initialised with max_attempts == 0).

        :param cache: The cache to store the downloaded file.
        :param transport: The transport to use for the API calls.
        :param max_workers: The maximum number of concurrent connections to use for data transfer. If None, the default
            is `min(32, os.cpu_count() + 4)`
        :param retry: A Retry object with a wait strategy. If None, a default Retry is created. If a chunk is successfully
            transferred the attempt counter will be reset.
        :param overwrite: whether to overwrite an existing local file
        :param fb: feedback to track the download, by tracking writes to the file

        :raises FileNameTooLongError: If the filename is too long.
        :raises ValueError: if the file already exists and overwrite is False
        :raises ChunkedIOException: if a non-recoverable exception occurred.
        :raises RetryError: if the maximum number of consecutive attempts have failed.
        """
        location = self._get_cache_location(cache)
        if not location.exists() or overwrite:
            await self.download_to_path(
                filename=location, transport=transport, max_workers=max_workers, retry=retry, fb=fb
            )
        else:
            logger.debug(f"Skipping download because data already in cache (label: {self.label})")
            fb.progress(1.0)
        return location



#  Copyright © 2025 Bentley Systems, Incorporated
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from __future__ import annotations

import asyncio
import contextlib
from collections.abc import Iterator
from typing import Any, Generic, TypeVar, cast
from uuid import UUID

import ipywidgets as widgets
from aiohttp.typedefs import StrOrURL
from IPython.display import display

from evo import logging
from evo.aio import AioTransport
from evo.common import APIConnector, BaseAPIClient, Environment
from evo.common.exceptions import UnauthorizedException
from evo.common.interfaces import IAuthorizer, ICache, IFeedback, ITransport
from evo.discovery import Hub, Organization
from evo.oauth import AnyScopes, EvoScopes, OAuthConnector
from evo.service_manager import ServiceManager
from evo.workspaces import Workspace

from ._consts import (
    DEFAULT_BASE_URI,
    DEFAULT_CACHE_LOCATION,
    DEFAULT_DISCOVERY_URL,
    DEFAULT_REDIRECT_URL,
)
from ._helpers import FileName, build_button_widget, build_img_widget, init_cache
from .authorizer import AuthorizationCodeAuthorizer
from .env import DotEnv

T = TypeVar("T")

logger = logging.getLogger(__name__)

__all__ = [
    "FeedbackWidget",
    "HubSelectorWidget",
    "OrgSelectorWidget",
    "ServiceManagerWidget",
    "WorkspaceSelectorWidget",
]


class DropdownSelectorWidget(widgets.HBox, Generic[T]):
    UNSELECTED: tuple[str, T]

    def __init__(self, label: str, env: DotEnv) -> None:
        self._env = env
        self.dropdown_widget = widgets.Dropdown(
            options=[self.UNSELECTED],
            description=label,
            value=self.UNSELECTED[1],
            layout=widgets.Layout(margin="5px 5px 5px 5px", align_self="flex-start"),
        )
        self.dropdown_widget.disabled = True
        self.dropdown_widget.observe(self._update_selected, names="value")

        self._loading_widget = build_img_widget("loading.gif")
        self._loading_widget.layout.display = "none"

        super().__init__([self.dropdown_widget, self._loading_widget])

    def _get_options(self) -> list[tuple[str, T]]:
        raise NotImplementedError("Subclasses must implement this method.")

    def _on_selected(self, value: T | None) -> None: ...

    @contextlib.contextmanager
    def _loading(self) -> Iterator[None]:
        self.dropdown_widget.disabled = True
        self._loading_widget.layout.display = "flex"
        try:
            yield
        finally:
            self._loading_widget.layout.display = "none"
            self.dropdown_widget.disabled = False

    def _update_selected(self, _: dict) -> None:
        self.selected = new_value = self.dropdown_widget.value
        self._on_selected(new_value if new_value != self.UNSELECTED[1] else None)

    def refresh(self) -> None:
        logger.debug(f"Refreshing {self.__class__.__name__} options...")
        self.dropdown_widget.disabled = True
        selected = self.selected
        self.dropdown_widget.options = options = [self.UNSELECTED] + self._get_options()
        if len(options) == 2 and selected == self.UNSELECTED[1]:
            # Automatically select the only option if there is only one and no missing option was previously selected.
            self.selected = new_value = options[1][1]
        else:
            # Otherwise, ensure the selected option is still valid.
            for _, value in options:
                if value == selected:
                    self.selected = new_value = selected
                    break
            else:
                # If the selected option is no longer valid, reset to the unselected value.
                self.selected = new_value = self.UNSELECTED[1]

        # Make sure the new value is passed to the _on_selected method.
        self._on_selected(new_value if new_value != self.UNSELECTED[1] else None)

        # Disable the widget if there are no options to select.
        self.dropdown_widget.disabled = len(options) <= 1

    @classmethod
    def _serialize(cls, value: T) -> str:
        raise NotImplementedError("Subclasses must implement this method.")

    @classmethod
    def _deserialize(cls, value: str) -> T:
        raise NotImplementedError("Subclasses must implement this method.")

    @property
    def selected(self) -> T:
        value = self._env.get(f"{self.__class__.__name__}.selected", self._serialize(self.UNSELECTED[1]))
        return self._deserialize(value)

    @selected.setter
    def selected(self, value: T) -> None:
        self._env.set(f"{self.__class__.__name__}.selected", self._serialize(value))
        self.dropdown_widget.value = value

    @property
    def disabled(self) -> bool:
        return self.dropdown_widget.disabled

    @disabled.setter
    def disabled(self, value: bool) -> None:
        self.dropdown_widget.disabled = value


_NULL_UUID = UUID(int=0)


class _UUIDSelectorWidget(DropdownSelectorWidget[UUID]):
    @classmethod
    def _serialize(cls, value: UUID) -> str:
        return str(value)

    @classmethod
    def _deserialize(cls, value: str) -> UUID:
        return UUID(value)


class OrgSelectorWidget(_UUIDSelectorWidget):
    UNSELECTED = ("Select Organisation", _NULL_UUID)

    def __init__(self, env: DotEnv, manager: ServiceManager) -> None:
        self._manager = manager
        super().__init__("Organisation", env)

    def _get_options(self) -> list[tuple[str, UUID]]:
        return [(org.display_name, org.id) for org in self._manager.list_organizations()]

    def _on_selected(self, value: UUID | None) -> None:
        self._manager.set_current_organization(value)


class HubSelectorWidget(DropdownSelectorWidget[str]):
    UNSELECTED = ("Select Hub", "")

    def __init__(self, env: DotEnv, manager: ServiceManager, org_selector: OrgSelectorWidget) -> None:
        self._manager = manager
        super().__init__("Hub", env)
        org_selector.dropdown_widget.observe(self._on_org_selected, names="value")

    def _on_org_selected(self, _: dict) -> None:
        self.refresh()

    def _get_options(self) -> list[tuple[str, str]]:
        return [(hub.display_name, hub.code) for hub in self._manager.list_hubs()]

    def _on_selected(self, value: str | None) -> None:
        self._manager.set_current_hub(value)

    @classmethod
    def _serialize(cls, value: str) -> str:
        return value

    @classmethod
    def _deserialize(cls, value: str) -> str:
        return value


class WorkspaceSelectorWidget(_UUIDSelectorWidget):
    UNSELECTED = ("Select Workspace", _NULL_UUID)

    def __init__(self, env: DotEnv, manager: ServiceManager, hub_selector: HubSelectorWidget) -> None:
        self._manager = manager
        super().__init__("Workspace", env)
        hub_selector.dropdown_widget.observe(self._on_hub_selected, names="value")

    async def refresh_workspaces(self) -> None:
        with self._loading():
            await self._manager.refresh_workspaces()
            self.refresh()

    def _on_hub_selected(self, _: dict) -> asyncio.Future:
        self.disabled = True
        return asyncio.ensure_future(self.refresh_workspaces())

    def _on_selected(self, value: UUID | None) -> None:
        self._manager.set_current_workspace(value)

    def _get_options(self) -> list[tuple[str, UUID]]:
        return [(ws.display_name, ws.id) for ws in self._manager.list_workspaces()]


# Generic type variable for the client factory method.
T_client = TypeVar("T_client", bound=BaseAPIClient)


class ServiceManagerWidget(widgets.HBox):
    def __init__(self, transport: ITransport, authorizer: IAuthorizer, discovery_url: str, cache: ICache) -> None:
        """
        :param transport: The transport to use for API requests.
        :param authorizer: The authorizer to use for API requests.
        :param discovery_url: The URL of the Evo Discovery service.
        :param cache: The cache to use for storing tokens and other data.
        """
        self._authorizer = authorizer
        self._service_manager = ServiceManager(
            transport=transport,
            authorizer=authorizer,
            discovery_url=discovery_url,
        )
        self._cache = cache
        env = DotEnv(cache)

        self._btn = build_button_widget("Sign In")
        self._btn.on_click(self._on_click)
        self._org_selector = OrgSelectorWidget(env, self._service_manager)
        self._hub_selector = HubSelectorWidget(env, self._service_manager, self._org_selector)
        self._workspace_selector = WorkspaceSelectorWidget(env, self._service_manager, self._hub_selector)

        self._loading_widget = build_img_widget("loading.gif")
        self._loading_widget.layout.display = "none"

        self._prompt_area = widgets.Output()
        self._prompt_area.layout.display = "none"

        col_1 = widgets.VBox(
            [
                widgets.HBox([build_img_widget("EvoBadgeCharcoal_FV.png"), self._btn, self._loading_widget]),
                widgets.HBox([self._org_selector]),
                widgets.HBox([self._hub_selector]),
                widgets.HBox([self._workspace_selector]),
            ]
        )
        col_2 = widgets.VBox([self._prompt_area])

        super().__init__(
            [col_1, col_2],
            layout={
                "display": "flex",
                "flex_flow": "row",
                "justify_content": "space-between",
                "align_items": "center",
            },
        )
        display(self)

    @classmethod
    def with_auth_code(
        cls,
        client_id: str,
        base_uri: str = DEFAULT_BASE_URI,
        discovery_url: str = DEFAULT_DISCOVERY_URL,
        redirect_url: str = DEFAULT_REDIRECT_URL,
        client_secret: str | None = None,
        cache_location: FileName = DEFAULT_CACHE_LOCATION,
        oauth_scopes: AnyScopes = EvoScopes.all_evo | EvoScopes.offline_access,
        proxy: StrOrURL | None = None,
    ) -> ServiceManagerWidget:
        """Create a ServiceManagerWidget with an authorization code authorizer.

        To use it, you will need an OAuth client ID. See the documentation for information on how to obtain this:
        https://developer.seequent.com/docs/guides/getting-started/apps-and-tokens

        Chain this method with the login method to authenticate the user and obtain an access token:

        ```python
        manager = await ServiceManagerWidget.with_auth_code(client_id="your-client-id").login()
        ```

        :param client_id: The client ID to use for authentication.
        :param base_uri: The OAuth server base URI.
        :param discovery_url: The URL of the Evo Discovery service.
        :param redirect_url: The local URL to redirect the user back to after authorisation.
        :param client_secret: The client secret to use for authentication.
        :param cache_location: The location of the cache file.
        :param oauth_scopes: The OAuth scopes to request.
        :param proxy: The proxy URL to use for API requests.

        :returns: The new ServiceManagerWidget.
        """
        cache = init_cache(cache_location)
        transport = AioTransport(user_agent=client_id, proxy=proxy)
        authorizer = AuthorizationCodeAuthorizer(
            oauth_connector=OAuthConnector(
                transport=transport,
                base_uri=base_uri,
                client_id=client_id,
                client_secret=client_secret,
            ),
            redirect_url=redirect_url,
            scopes=oauth_scopes,
            env=DotEnv(cache),
        )
        return cls(transport, authorizer, discovery_url, cache)

    async def _login_with_auth_code(self, timeout_seconds: int) -> None:
        """Login using an authorization code authorizer.

        This method will attempt to reuse an existing token from the environment file. If no token is found, the user will
        be prompted to log in.

        :param timeout_seconds: The number of seconds to wait for the user to log in.
        """
        authorizer = cast(AuthorizationCodeAuthorizer, self._authorizer)
        if not await authorizer.reuse_token():
            await authorizer.login(timeout_seconds=timeout_seconds)

    async def login(self, timeout_seconds: int = 180) -> ServiceManagerWidget:
        """Authenticate the user and obtain an access token.

        Only the notebook authorizer implementations are supported by this method.

        This method returns the current instance of the ServiceManagerWidget to allow for method chaining.

        ```python
        manager = await ServiceManagerWidget.with_auth_code(client_id="your-client-id").login()
        ```

        :param timeout_seconds: The maximum time (in seconds) to wait for the authorisation process to complete.

        :returns: The current instance of the ServiceManagerWidget.
        """
        # Open the transport without closing it to avoid the overhead of opening it multiple times.
        await self._service_manager._transport.open()
        with self._loading():
            match self._authorizer:
                case AuthorizationCodeAuthorizer():
                    await self._login_with_auth_code(timeout_seconds)
                case unknown:
                    raise NotImplementedError(f"ServiceManagerWidget cannot login using {type(unknown).__name__}.")

            # Refresh the services after logging in.
            await self.refresh_services()
        return self

    @property
    def cache(self) -> ICache:
        return self._cache

    def _update_btn(self, signed_in: bool) -> None:
        if signed_in:
            self._btn.description = "Refresh Evo Services"
        else:
            self._btn.description = "Sign In"

    def _on_click(self, _: widgets.Button) -> asyncio.Future:
        return asyncio.ensure_future(self.refresh_services())

    @contextlib.contextmanager
    def _loading(self) -> Iterator[None]:
        self._btn.disabled = True
        self._loading_widget.layout.display = "flex"
        try:
            yield
        finally:
            self._loading_widget.layout.display = "none"
            self._btn.disabled = False

    @contextlib.contextmanager
    def _loading_services(self) -> Iterator[None]:
        self._org_selector.disabled = True
        self._hub_selector.disabled = True
        self._workspace_selector.disabled = True
        try:
            yield
        finally:
            self._org_selector.refresh()
            self._hub_selector.refresh()

    @contextlib.contextmanager
    def _prompt(self) -> Iterator[widgets.Output]:
        self._prompt_area.layout.display = "flex"
        try:
            yield self._prompt_area
        finally:
            self._prompt_area.layout.display = "none"
            self._prompt_area.clear_output()

    async def refresh_services(self) -> None:
        with self._loading():
            with self._loading_services():
                try:
                    await self._service_manager.refresh_organizations()
                except UnauthorizedException:  # Expired token or user not logged in.
                    # Attempt to log in again.
                    await self.login()

                    # Try refresh the services again after logging in.
                    await self._service_manager.refresh_organizations()
            await self._workspace_selector.refresh_workspaces()
            self._update_btn(True)

    @property
    def organizations(self) -> list[Organization]:
        return self._service_manager.list_organizations()

    @property
    def hubs(self) -> list[Hub]:
        return self._service_manager.list_hubs()

    @property
    def workspaces(self) -> list[Workspace]:
        return self._service_manager.list_workspaces()

    def get_connector(self) -> APIConnector:
        """Get an API connector for the currently selected hub.

        :returns: The API connector.

        :raises SelectionError: If no organization or hub is currently selected.
        """
        return self._service_manager.get_connector()

    def get_environment(self) -> Environment:
        """Get an environment with the currently selected organization, hub, and workspace.

        :returns: The environment.

        :raises SelectionError: If no organization, hub, or workspace is currently selected.
        """
        return self._service_manager.get_environment()

    def create_client(self, client_class: type[T_client], *args: Any, **kwargs: Any) -> T_client:
        """Create a client for the currently selected workspace.

        :param client_class: The class of the client to create.

        :returns: The new client.

        :raises SelectionError: If no organization, hub, or workspace is currently selected.
        """
        return self._service_manager.create_client(client_class, *args, **kwargs)


class FeedbackWidget(IFeedback):
    """Simple feedback widget for displaying progress and messages to the user."""

    def __init__(self, label: str) -> None:
        """
        :param label: The label for the feedback widget.
        """
        label = widgets.Label(label)
        self._progress = widgets.FloatProgress(value=0, min=0, max=1, style={"bar_color": "#265C7F"})
        self._progress.layout.width = "400px"
        self._msg = widgets.Label("", style={"font_style": "italic"})
        self._widget = widgets.HBox([label, self._progress, self._msg])
        self._last_message = ""
        display(self._widget)

    def progress(self, progress: float, message: str | None = None) -> None:
        """Progress the feedback and update the text to message.

        This can raise an exception to cancel the current operation.

        :param progress: A float between 0 and 1 representing the progress of the operation as a percentage.
        :param message: An optional message to display to the user.
        """
        self._progress.value = progress
        self._progress.description = f"{progress * 100:5.1f}%"
        if message is not None:
            self._msg.value = message



#  Copyright © 2025 Bentley Systems, Incorporated
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from __future__ import annotations

import asyncio
import hashlib
import secrets
import webbrowser
from base64 import urlsafe_b64encode
from types import TracebackType
from urllib.parse import urlencode, urlparse

from evo import logging

from .connector import OAuthConnector
from .data import AccessToken, AnyScopes
from .exceptions import OAuthError

try:
    from aiohttp import web
except ImportError:
    raise ImportError(f"{__name__} could not be used because OAuth dependencies are not installed.")

logger = logging.getLogger("oauth")

__all__ = ["OAuthRedirectHandler"]


class OAuthRedirectHandler:
    """An asynchronous context manager for handling OAuth authorisation redirects. Not thread-safe.

    This context manager starts an HTTP server to handle the OAuth redirect from the server. The user should be
    redirected to the local server to complete the authorisation process. The context manager waits for the
    authorisation to complete and then fetches the access token.
    """

    _REDIRECT_HTML = """
<html>
  <body>
    <h1>Authorization request to Seequent has been completed.</h1>
    <p>You may close this tab or window now.</p>
    <script>setTimeout("window.close()", 2500);</script>
  </body>
</html>
""".encode("UTF-8")

    def __init__(self, oauth_connector: OAuthConnector, redirect_url: str) -> None:
        """
        :param oauth_connector: The OAuth connector to use for authorisation.
        """
        self.__oauth_connector = oauth_connector
        self.__runner: web.AppRunner | None = None  # The HTTP server runner.
        self.__authorisation: asyncio.Future[AccessToken] = asyncio.get_event_loop().create_future()
        self.__redirect_url = redirect_url
        self.__state: str | None = None
        self.__verifier: str | None = None

    @property
    def pending(self) -> bool:
        """Check if the context is still pending.

        True if the context is still pending, False otherwise.

        The context is pending if the token has not been fetched and no errors have occurred.
        """
        return not self.__authorisation.done()

    def __check_server_started(self) -> None:
        """Check if the OAuth server has been started.

        :raises OAuthError: If the server has not been started.
        """
        if self.__runner is None:
            raise OAuthError("OAuth HTTP server not started.")

    async def __aenter__(self) -> OAuthRedirectHandler:
        """Start the OAuth HTTP server."""
        if self.__runner is not None:
            raise OAuthError("OAuth redirect server cannot be reused.")

        logger.debug("Opening connector...")
        await self.__oauth_connector.__aenter__()

        logger.debug("Configuring OAuth HTTP server...")
        app = web.Application(logger=logger)
        uri = urlparse(self.__redirect_url)
        app.add_routes([web.get(uri.path, self.__handle_request)])
        self.__runner = runner = web.AppRunner(app)
        await runner.setup()

        logger.debug("Starting OAuth HTTP server...")
        await web.TCPSite(runner, uri.hostname, uri.port, ssl_context=None).start()

        logger.debug("OAuth HTTP server started.")
        return self

    async def __aexit__(
        self, exc_type: type[Exception] | None, exc_val: Exception | None, exc_tb: TracebackType | None
    ) -> None:
        """Stop the OAuth HTTP server."""
        self.__check_server_started()

        logger.debug("Stopping OAuth HTTP server...")
        await self.__runner.cleanup()
        logger.debug("OAuth HTTP server stopped.")

        logger.debug("Closing connector...")
        await self.__oauth_connector.__aexit__(exc_type, exc_val, exc_tb)

    async def __handle_request(self, request: web.Request) -> web.StreamResponse:
        """Handle the in-browser redirect from the OAuth server.

        This method is called when the user is redirected back to the local server.

        Errors that are encountered by this handler would usually be invisible to other parts of the application, so
        they are logged and stored in the context for later retrieval.
        """
        # This request is already successful. Any further errors will be raised in application code.
        response = web.StreamResponse(status=200, headers={"Content-Type": "text/html"})
        await response.prepare(request)
        await response.write(self._REDIRECT_HTML)
        try:  # Broad exception handling to ensure any errors are logged and stored in the context.
            if "error" in request.query:  # Check for an error response from the OAuth provider.
                title = request.query.getone("error")  # Raises KeyError if `error` is missing.
                detail = request.query.getone("error_description", None)
                raise OAuthError(detail or title)  # Report the more detailed error message if it is available.

            token = await self.get_token(
                request.query.getone("state"),  # Raises KeyError if `state` is missing.
                request.query.getone("code"),  # Raises KeyError if `code` is missing.
            )
            self.__authorisation.set_result(token)
        except Exception as exc:  # noqa: E722
            logger.error("Unable to fetch access token.", exc_info=True)
            self.__authorisation.set_exception(OAuthError(str(exc)).with_traceback(exc.__traceback__))
        finally:  # Always return the response to the web client.
            return response

    async def get_result(self, timeout_seconds: int | float = 60) -> AccessToken:
        """Get the result of the authorisation process.

        This method will block until the authorisation process is complete or the timeout is reached.

        :param timeout_seconds: The maximum time (in seconds) to wait for the authorisation process to complete.

        :return: The access token.

        :raises OAuthError: If the authorisation process times out.
        :raises OAuthError: If an error occurred during the authorisation process.
        """
        self.__check_server_started()
        try:
            return await asyncio.wait_for(self.__authorisation, timeout_seconds)
        except asyncio.TimeoutError:
            raise OAuthError("Timed out waiting for OAuth response.")

    async def login(self, scopes: AnyScopes, timeout_seconds: int | float = 60) -> AccessToken:
        """Authenticate the user and obtain an access token.

        This method will launch a web browser to authenticate the user and obtain an access token.

        :param scopes: The OAuth scopes to request.
        :param timeout_seconds: The maximum time (in seconds) to wait for the authorisation process to complete.

        :return: The access token.

        :raises OAuthError: If the user does not authenticate within the timeout.
        :raises OAuthError: If an error occurred during the authorisation process.
        """
        self.__check_server_started()
        auth_uri = self.create_authorization_url(scopes)
        webbrowser.open(auth_uri)
        return await self.get_result(timeout_seconds)

    def create_authorization_url(self, scopes: AnyScopes) -> str:
        """Create an authorization URL for the given client.

        https://www.oauth.com/oauth2-servers/pkce/authorization-request/

        :param scopes: The scopes to request.

        :return: The authorization URL.
        """
        base_url = self.__oauth_connector.base_uri + self.__oauth_connector.endpoint("authorize")
        challenge, state = self._get_challenge_and_state()
        qs_params = {  # The query string parameters for the authorisation URL.
            "response_type": "code",
            "client_id": self.__oauth_connector.client_id,
            "redirect_uri": self.__redirect_url,
            "scope": str(scopes),
            "state": state,
            "code_challenge": challenge,
            "code_challenge_method": "S256",
        }
        return base_url + "?" + urlencode(sorted(qs_params.items()))

    def _get_challenge_and_state(self) -> tuple[str, str]:
        """Generate the PKCE challenge and state for the OAuth request.

        https://www.oauth.com/oauth2-servers/pkce/authorization-request/

        A new state, code verifier, and PKCE challenge are generated each time this method is called to ensure
        that the challenge is unique for each attempt.

        :return: A tuple containing the encoded PKCE challenge and the state.
        """
        self.__state = state = secrets.token_urlsafe(32)  # A unique state to prevent CSRF attacks.
        self.__verifier = verifier = secrets.token_urlsafe(48)  # A unique code verifier for the PKCE challenge.
        code_verifier = verifier.encode("utf-8")
        digest = hashlib.sha256(code_verifier).digest()
        return urlsafe_b64encode(digest).decode("utf-8").rstrip("="), state

    async def get_token(self, state: str, code: str) -> AccessToken:
        """Get an access token from the server.

        https://www.oauth.com/oauth2-servers/access-tokens/authorization-code-request/

        This method should be called by the client's redirect handler with the parameters from the
        authorisation response.

        :param state: The state parameter from the authorisation response.
        :param code: The code parameter from the authorisation response.

        :return: The access token.

        :raises OAuthError: If the state is invalid or the token cannot be fetched.
        """
        logger.debug("Validating state...")
        if state != self.__state:  # The state must match the one generated by this client.
            raise OAuthError("Invalid state.")

        data = {  # The payload to send to the OAuth server in the token request.
            "grant_type": "authorization_code",
            "code": code,
            "code_verifier": self.__verifier,
            "redirect_uri": self.__redirect_url,
        }

        logger.debug("Fetching access token...")
        token = await self.__oauth_connector.fetch_token(data, AccessToken)

        return token



#  Copyright © 2025 Bentley Systems, Incorporated
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import functools
from importlib import metadata

__all__ = ["get_package_details"]


@functools.cache
def get_package_details(candidate: str) -> dict:
    """Get package name and version given the module __name__.

    :param candidate: The module __name__ to start searching from.

    :return: A dictionary containing the package name and version.

    :raises metadata.PackageNotFoundError: If no package could be found.
    """
    while candidate:
        try:
            package_metadata = metadata.metadata(candidate)
        except metadata.PackageNotFoundError:
            candidate, *_ = candidate.rpartition(".")
        else:
            return {
                "name": package_metadata["name"],
                "version": package_metadata["version"],
            }

    raise metadata.PackageNotFoundError(__package__)
